#include <iostream>
#include <fstream>
#include <NvInferPlugin.h>
#include <NvOnnxParser.h>
#include <NvInfer.h>
#include <cuda_runtime.h>
#include <cassert>
#include <memory>

class TrtLogger : public nvinfer1::ILogger {
  void log(Severity severity, const char* msg) noexcept override {
    if (severity != Severity::kVERBOSE && msg && msg[0]) {
      // noexcept implementation
      printf("%s\n", msg);
    }
  }
};

void buildEngine() {
  std::cout << "Building model" << std::endl;
  std::string onnx_filename = "./model.onnx";
  std::string trt_filename = "./model.trt";
  TrtLogger g_logger_{};

  nvinfer1::IBuilder *builder_ = nvinfer1::createInferBuilder(g_logger_);
  builder_->setMaxBatchSize(2);

  nvinfer1::IBuilderConfig *builder_config_ = builder_->createBuilderConfig();
  builder_config_->setMaxWorkspaceSize(1.5 * (1 << 30));
  builder_config_->setFlag(nvinfer1::BuilderFlag::kGPU_FALLBACK);
  builder_config_->setDefaultDeviceType(nvinfer1::DeviceType::kGPU);
  //  builder_config_->setFlag(nvinfer1::BuilderFlag::kFP16);

  std::ifstream onnx_file(onnx_filename.c_str(), std::ifstream::binary);
  auto flag = nvinfer1::NetworkDefinitionCreationFlag::kEXPLICIT_BATCH;
  auto explicitBatch = 1U << static_cast<uint32_t>(flag);
  nvinfer1::INetworkDefinition *network_ = builder_->createNetworkV2(explicitBatch);
  auto parser_ = nvonnxparser::createParser(*network_, g_logger_);
  parser_->parseFromFile(onnx_filename.c_str(),
                         static_cast<int>(nvinfer1::ILogger::Severity::kWARNING));
  for (int i = 0; i < parser_->getNbErrors(); ++i) {
    std::cerr << "ERROR: " << parser_->getError(i)->desc() << std::endl;
  }

  auto profile = builder_->createOptimizationProfile();
  profile->setDimensions(network_->getInput(0)->getName(),
                         nvinfer1::OptProfileSelector::kMAX,
                         nvinfer1::Dims4{2, 3, 20 , 40});
  builder_config_->addOptimizationProfile(profile);


  auto engine_memory_ = builder_->buildSerializedNetwork(*network_,
                                                         *builder_config_);
  assert(engine_memory_ != nullptr);

  std::ofstream engine_file;
  engine_file.open(trt_filename, std::ios::binary | std::ios::out);
  engine_file.write((const char*)engine_memory_->data(), engine_memory_->size());
  engine_file.close();
  onnx_file.close();
  std::cout << "Building model has completed " << std::endl;
}

void loadModel(){
  std::cout << "Loading model" << std::endl;
  TrtLogger g_logger_{};
  std::string trt_filename = "./model.trt";
  std::ifstream engine_file(trt_filename.c_str(), std::ifstream::binary);
  auto const start_pos = engine_file.tellg();
  engine_file.ignore(std::numeric_limits<std::streamsize>::max());
  size_t buffer_count = engine_file.gcount();

  engine_file.seekg(start_pos);
  std::unique_ptr<char[]> engine_buffer(new char[buffer_count]);
  engine_file.read(engine_buffer.get(), buffer_count);
  auto runtime_ = nvinfer1::createInferRuntime(g_logger_);
  assert(runtime_ != nullptr);
  auto engine_ = runtime_->deserializeCudaEngine((void*)engine_buffer.get(),
                                                 buffer_count);
  assert(engine_ != nullptr);
  auto context_ = engine_->createExecutionContext();
  assert(context_ != nullptr);
  engine_file.close();
  std::cout << "================= Engine Info =================" << std::endl;
  auto batch_size_engine_ = engine_->getMaxBatchSize();
  std::cout << "batch_size_engine : " << batch_size_engine_ << std::endl;
  auto num_bindings_ = engine_->getNbBindings();
  std::cout << "num_bindings : " << num_bindings_ << std::endl;

  for (int i = 0; i < num_bindings_; i++) {
    nvinfer1::Dims dims = engine_->getBindingDimensions(i);
    const char* name = engine_->getBindingName(i);
    if (engine_->bindingIsInput(i)) {
      std::cout << "================= Input =================" << std::endl;

    } else {
      std::cout << "================= Output =================" << std::endl;
    }
    std::cout << "\tName : " << name << std::endl;
    std::cout << "\tShape : [";
    for (int j = 0; j<dims.nbDims; j++){
      std::cout << dims.d[j] << ", ";
    }
    std::cout << "]" << std::endl;
  }

  std::cout << "================= Context Info =================" << std::endl;
  std::cout << "Dimension Specified " << context_->allInputDimensionsSpecified()
                                      << std::endl;
  auto input_dims = context_->getBindingDimensions(0);
  std::cout << "\tInput Shape : [";
  for (int j = 0; j<input_dims.nbDims; j++){
    std::cout << input_dims.d[j] << ", ";
  }
  std::cout << "]" << std::endl;

  auto output_dims = context_->getBindingDimensions(1);
  std::cout << "\tOutput Shape : [";
  for (int j = 0; j<output_dims.nbDims; j++){
    std::cout << output_dims.d[j] << ", ";
  }
  std::cout << "]" << std::endl;

  std::cout << "================= After setting binding dimension to Context "
               "Info =================" << std::endl;
  context_->setBindingDimensions(0,
                                 nvinfer1::Dims4(2, 3, 20, 40));
  output_dims = context_->getBindingDimensions(0);
  std::cout << "\tOutput Shape : [";
  for (int j = 0; j<output_dims.nbDims; j++){
    std::cout << output_dims.d[j] << ", ";
  }
  std::cout << "]" << std::endl;

}


int main(int argc, char** argv) {
  bool build = std::atoi(argv[1]);
  if (build == 0){
    buildEngine();
  } else {
    loadModel();
  }
  return 0;
}
